* wwDotNetBridge.DataSetToCursors converts a DataSet into cursors but it's too
* slow when there are a lot of records because it converts the DataSet to XML
* and then uses XMLAdapter to create a cursor, and XMLAdapter is really slow
* when there are lots of records. So, we'll take a different approach: create a
* cursor with the same structure as the specified DataTable and fill the cursor
* with the contents of each row.

lparameters toBridge, ;
	toTable, ;
	tcCursor, ;
	tnRecords
local lnColumns, ;
	lcNull, ;
	lcCursor, ;
	laColumns[1], ;
	laMaxLength[1], ;
	lnI, ;
	loColumn, ;
	lcColumnName, ;
	lcDataType, ;
	lcType, ;
	lcAlias, ;
	lnRecords, ;
	lnRows, ;
	loRow, ;
	lnJ, ;
	luValue, ;
	lcError, ;
	laFields[1], ;
	lnFields, ;
	lnV, ;
	lcVrbl, ;
	lcFType, ;
	llSame, ;
	llError

* Figure out how many columns there are.

lnColumns = min(toBridge.GetPropertyEx(toTable, 'Columns.Count'), 254)
if lnColumns = 0
	return .F.
endif lnColumns = 0

* Store each column name in an array and create a CREATE CURSOR statement.

lcNull = set('NULL')
set null on
lcCursor = ''
dimension laColumns[lnColumns, 3], laMaxLength[lnColumns]
laMaxLength = 0
for lnI = 0 to lnColumns - 1
	loColumn     = toBridge.GetPropertyEx(toTable, 'Columns[' + ;
		transform(lnI) + ']')
	lcColumnName = GetValidName(loColumn.ColumnName)
	lcDataType   = toBridge.GetPropertyEx(loColumn.DataType, 'Name')
	do case
		case inlist(lcDataType, 'Decimal', 'Double', 'Single')
			lcType = 'B(8)'
		case lcDataType = 'Boolean'
			lcType = 'L'
		case lcDataType = 'DateTime'
			lcType = 'T'
		case lcDataType = 'Byte[]'
			lcType = 'W'
		case inlist(lcDataType, 'Int', 'Byte')
			lcType = 'I'
		case lcDataType = 'Guid'
			lcType = 'C(36)'
		case loColumn.MaxLength = -1
			lcType = 'M'
		otherwise
			lcType = 'C(' + transform(loColumn.MaxLength) + ')'
	endcase
	if inlist(upper(lcColumnName) + ' ', 'NULL ', 'NOT ', 'CHECK ', 'ERROR ', ;
		'AUTOINC ', 'NEXTVALUE ', 'STEP ', 'DEFAULT ', 'PRIMARY ', 'KEY ', ;
		'UNIQUE ', 'COLLATE ', 'REFERENCES ', 'TAG ', 'NOCPTRANS ')
		lcColumnName = '[' + lcColumnName + ']'
	endif inlist(upper(lcColumnName) ...
	lcCursor = lcCursor + iif(empty(lcCursor), '', ',') + lcColumnName + ;
		' ' + lcType
	laColumns[lnI + 1, 1] = lcColumnName
	laColumns[lnI + 1, 2] = left(lcType, 1)
	laColumns[lnI + 1, 3] = lcDataType
	local &lcColumnName
	if len(lcCursor) > 8000
		exit
	endif len(lcCursor) > 8000
next lnI

* Create the cursor.

lcAlias  = sys(2015)
lcCursor = 'create cursor ' + lcAlias + ' (' + lcCursor + ')'
&lcCursor

* Go through each row, get each column, and populate the cursor.

lnRecords = evl(tnRecords, 9999999999)
lnRows    = toBridge.GetPropertyEx(toTable, 'Rows.Count')
for lnI = 0 to min(lnRows, lnRecords) - 1
	loRow = toBridge.GetPropertyEx(toTable, 'Rows[' + transform(lnI) + ']')
	for lnJ = 0 to lnColumns - 1
		luValue = toBridge.GetPropertyEx(loRow, 'ItemArray[' + ;
			transform(lnJ) + ']')
		do case
			case inlist(laColumns[lnJ + 1, 3], 'Object', 'Type')
				luValue = toBridge.InvokeMethod(luValue, 'ToString')
				laMaxLength[lnJ + 1] = max(laMaxLength[lnJ + 1], len(luValue))
			case laColumns[lnJ + 1, 3] = 'Guid'
				luValue = luValue.GuidString
				laMaxLength[lnJ + 1] = 36
			case vartype(luValue) = 'C' or laColumns[lnJ + 1, 2] $ 'CM'
				luValue = transform(luValue)
				laMaxLength[lnJ + 1] = max(laMaxLength[lnJ + 1], len(luValue))
		endcase
		store luValue to (laColumns[lnJ + 1, 1])
	next lnJ
	try
		insert into (lcAlias) from memvar
	catch

* If we have a data type issue, indicate which column.

		lcError  = ''
		lnFields = afields(laFields)
		for lnV = 1 to lnFields
			lcVrbl  = 'M.' + laFields[lnV, 1]
			lcFType = laFields[lnV, 2]
			lcType  = type(lcVrbl)
			llSame  = lcType = lcFType or ;
				(lcType = 'N' and lcFType $ 'NFIBY') or ;
				(lcType $ 'CMV' and lcFType $ 'CMV') or ;
				empty(lcType)
			if not llSame
				lcError = lcError + lcVrbl + ' Variable type: ' + lcType + ;
					' Field type: ' + lcFType + chr(13)
			endif not llSame
		next lnV
		llError = .T.
	endtry
	if llError
		exit
	endif llError
next lnI

* Do a final select to get the correct column lengths.

if not llError
	lcCursor = ''
	for lnI = 1 to alen(laColumns, 1)
		lcColumnName = laColumns[lnI, 1]
		lcType       = laColumns[lnI, 2]
		if lcType $ 'CM' and laMaxLength[lnI] < 255
			lcType = 'C(' + transform(max(laMaxLength[lnI], 1)) + ')'
		endif lcType $ 'CM' ...
		lcCursor = lcCursor + iif(empty(lcCursor), '', ',') + lcColumnName + ;
			' ' + lcType
		if len(lcCursor) > 8000
			exit
		endif len(lcCursor) > 8000
	next lnI
	lcCursor = 'create cursor ' + tcCursor + ' (' + lcCursor + ')'
	&lcCursor
	append from dbf(lcAlias)
	go top
endif not llError
if lcNull = 'OFF'
	set null off
endif lcNull = 'OFF'
use in (lcAlias)
if llError
	error lcError
endif llError
return not llError


function GetValidName(tcName)
local lcIllegal, ;
	lcDelimiters, ;
	lcName

* Decide what are illegal characters and the usual suspects for field
* delimiters.

lcIllegal    = [ ~!@#$%^&*()-+=-/?{}|<>,:;\.']
lcDelimiters = '"[]`'

* Convert illegal characters to _ and strip out delimiters.

lcName = chrtran(tcName, lcIllegal, replicate('_', len(lcIllegal)))
lcName = chrtran(lcName, lcDelimiters, '')
lcName = iif(isdigit(left(lcName, 1)), 'A', '') + lcName
return lcName
