*==============================================================================
* Method:			DecryptString
* Purpose:			Decrypt a string
* Author:			Doug Hennig
* Last Revision:	10/05/2017
* Parameters:		tcString - the string to decrypt
*					tcKey    - a key to decrypt against
* Returns:			the decrypted string
* Environment in:	oBridge contains an instance of wwDotNetBridge
*					Encryption.dll has been loaded
*					VFPEncryption.fll is available
* Environment out:	VFPEncryption.fll may be open as a library file
*==============================================================================

lparameters tcString, ;
	tcKey
local lcString

* Handle a blank string or key.

if empty(tcString) or empty(tcKey)
	return tcString
endif empty(tcString) ...

* If the right-most character of the encrypted string is CHR(1), the string
* was encrypted using Encryption.dll, so use that to decrypt it.

lcString = trim(tcString)
if right(lcString, 1) = chr(1)
	lcString = left(lcString, len(lcString) - 1)
	return nvl(oBridge.InvokeStaticMethod('Encryption.Encryption', 'Decrypt', ;
		lcString, tcKey, chr(102) + chr(57) + chr(110) + chr(73) + chr(73) + ;
		chr(71) + chr(97) + chr(89) + chr(70) + chr(51) + chr(54) + chr(97) + ;
		chr(69) + chr(55) + chr(48) + chr(68)), '')

* The string was encrypted using VFPEncryption.fll, so use that to decrypt it.

else
	if not 'vfpencryption.fll' $ lower(set('LIBRARY'))
		lcLibrary = 'VFPEncryption.fll'
		set library to (lcLibrary) additive
	endif not 'vfpencryption.fll' $ lower(set('LIBRARY'))
	return Decrypt(lcString, tcKey, 1024)
endif right(lcString, 1) = chr(1)
